##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2014 Stanford University and the Authors
#
# Authors: Kyle A. Beauchamp
# Contributors: Robert McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
#############################################################################

from __future__ import print_function, absolute_import

import os
import subprocess
from distutils.spawn import find_executable as _find_executable
from distutils.version import LooseVersion

import numpy as np

from mdtraj.utils import enter_temp_directory, import_

##############################################################################
# Globals
##############################################################################

# Possible names for the external commands -- these are expected
# to be found in the PATH.
SHIFTX2 = ['shiftx2.py']
SPARTA_PLUS = ['sparta+', 'SPARTA+', 'SPARTA+.linux']
PPM = ['ppm_linux_64.exe']

__all__ = ['chemical_shifts_shiftx2', 'chemical_shifts_ppm', 'chemical_shifts_spartaplus', "reindex_dataframe_by_atoms"]


def find_executable(names):
    for possible in names:
        result = _find_executable(possible)
        if result is not None:
            return result
    return None


##############################################################################
# Code
##############################################################################

def compute_chemical_shifts(trj, model="shiftx2", **kwargs):
    """Predict chemical shifts of a trajectory using ShiftX2.

    Parameters
    ----------
    trj : Trajectory
        Trajectory to predict shifts for.
    model : str, optional, default="shiftx2"
        The program to use for calculating chemical shifts.  Must be one
        of shiftx2, ppm, or sparta+

    Returns
    -------
    results : pandas DataFrame
        Dataframe containing results, with index consisting of
        (resSeq, atom_name) pairs and columns for each frame in trj.

    Notes
    -----
    You must have the appropriate chemical soft programs installed
    and in your executable path.

    Chemical shift prediction is for PROTEIN atoms; trajectory objects
    with ligands, solvent, ions, or other non-protein components may give
    UNKNOWN RESULTS.

    Please cite the appropriate reference, see docstrings for chemical_shifts_*
    for the various possible models.
    """
    if model == "shiftx2":
        return chemical_shifts_shiftx2(trj, **kwargs)
    elif model == "ppm":
        return chemical_shifts_ppm(trj, **kwargs)
    elif model == "sparta+":
        return chemical_shifts_spartaplus(trj, **kwargs)
    else:
        raise(ValueError("model must be one of shiftx2, ppm, or sparta+"))


def chemical_shifts_shiftx2(trj, pH=5.0, temperature=298.00):
    """Predict chemical shifts of a trajectory using ShiftX2.

    Parameters
    ----------
    trj : Trajectory
        Trajectory to predict shifts for.
    pH : float, optional, default=5.0
        pH value which gets passed to the ShiftX2 predictor.
    temperature : float, optional, default=298.00
        Temperature which gets passed to the ShiftX2 predictor.

    Returns
    -------
    results : pandas DataFrame
        Dataframe containing results, with index consisting of
        (resSeq, atom_name) pairs and columns for each frame in trj.

    Notes
    -----
    You must have ShiftX2 available on your path; see (http://www.shiftx2.ca/).

    Chemical shift prediction is for PROTEIN atoms; trajectory objects
    with ligands, solvent, ions, or other non-protein components may give
    UNKNOWN RESULTS.

    Please cite the appropriate reference below.

    References
    ----------
    .. [1] Beomsoo Han, Yifeng Liu, Simon Ginzinger, and David Wishart.
       "SHIFTX2: significantly improved protein chemical shift
       prediction." J. Biomol. NMR, 50, 1 43-57 (2011)
    """
    pd = import_('pandas')
    binary = find_executable(SHIFTX2)
    if binary is None:
        raise OSError('External command not found. Looked for {} in PATH. '
                      '`chemical_shifts_shiftx2` requires the external program SHIFTX2, '
                      'available at http://www.shiftx2.ca/'.format(', '.join(SHIFTX2)))

    results = []
    with enter_temp_directory():
        for i in range(trj.n_frames):
            fn = './trj%d.pdb' % i
            trj[i].save(fn)
            subprocess.check_call([binary,
                                   '-b', fn,
                                   '-p', "{:.1f}".format(pH),
                                   '-t', "{:.2f}".format(temperature),
                                   ])

            d = pd.read_csv("./trj%d.pdb.cs" % i)
            d.rename(columns={"NUM": "resSeq", "RES": "resName", "ATOMNAME": "name"}, inplace=True)
            d["frame"] = i
            results.append(d)

    results = pd.concat(results)

    if LooseVersion(pd.__version__) < LooseVersion('0.14.0'):
        results = results.pivot_table(rows=["resSeq", "name"], cols="frame", values="SHIFT")
    else:
        results = results.pivot_table(index=["resSeq", "name"], columns="frame", values="SHIFT")

    return results


def chemical_shifts_ppm(trj):
    """Predict chemical shifts of a trajectory using ppm.

    Parameters
    ----------
    trj : Trajectory
        Trajectory to predict shifts for.

    Returns
    -------
    results : pandas.DataFrame
        Dataframe containing results, with index consisting of
        (resSeq, atom_name) pairs and columns for each frame in trj.

    Notes
    -----
    You must have ppm available on your path; see
    (http://spin.ccic.ohio-state.edu/index.php/download/index).

    Chemical shift prediction is for PROTEIN atoms; trajectory objects
    with ligands, solvent, ions, or other non-protein components may give
    UNKNOWN RESULTS.

    Please cite the appropriate reference below.

    References
    ----------
    .. [1] Li, DW, and Bruschweiler, R. "PPM: a side-chain and backbone chemical
       shift predictor for the assessment of protein conformational ensembles."
       J Biomol NMR. 2012 Nov;54(3):257-65.
    """
    pd = import_('pandas')
    binary = find_executable(PPM)

    first_resSeq = trj.top.residue(0).resSeq

    if binary is None:
        raise OSError('External command not found. Looked for %s in PATH. `chemical_shifts_ppm` requires the external program PPM, available at http://spin.ccic.ohio-state.edu/index.php/download/index' % ', '.join(PPM))

    with enter_temp_directory():
        trj.save("./trj.pdb")
        #-para old is on order to use newer ppm versions
        cmd = "%s -pdb trj.pdb -mode detail -para old" % binary

        return_flag = os.system(cmd)

        if return_flag != 0:
            raise(IOError("Could not successfully execute command '%s', check your PPM installation or your input trajectory." % cmd))

        d = pd.read_table("./bb_details.dat", index_col=False, header=None, sep="\s+").drop([3], axis=1)

        d = d.rename(columns={0: "resSeq", 1: "resName", 2: "name"})
        d["resSeq"] += first_resSeq - 1  # Fix bug in PPM that reindexes to 1
        d = d.drop("resName", axis=1)
        d = d.set_index(["resSeq", "name"])
        d.columns = np.arange(trj.n_frames)
        d.columns.name = "frame"

    return d


def _get_lines_to_skip(filename):
    """Determine the number of comment lines in a SPARTA+ output file."""
    format_string = """FORMAT %4d %4s %4s %9.3f %9.3f %9.3f %9.3f %9.3f %9.3f"""
    handle = open(filename)
    for i, line in enumerate(handle):
        if line.find(format_string) != -1:
            return i + 2

    raise(Exception("No format string found in SPARTA+ file!"))


def chemical_shifts_spartaplus(trj, rename_HN=True):
    """Predict chemical shifts of a trajectory using SPARTA+.

    Parameters
    ----------
    trj : Trajectory
        Trajectory to predict shifts for.
    rename_HN : bool, optional, default=True
        SPARTA+ calls the amide proton "HN" instead of the standard "H".
        When True, this option renames the output as "H" to match the PDB
        and BMRB nomenclature.

    Returns
    -------
    results : pandas.DataFrame
        Dataframe containing results, with index consisting of
        (resSeq, atom_name) pairs and columns for each frame in trj.

    Notes
    -----
    You must have SPARTA+ available on your path; see
    (http://spin.niddk.nih.gov/bax/software/SPARTA+/). Also, the SPARTAP_DIR
    environment variable must be set so that SPARTA+ knows where to find
    its database files.

    Chemical shift prediction is for PROTEIN atoms; trajectory objects
    with ligands, solvent, ions, or other non-protein components may give
    UNKNOWN RESULTS.

    Please cite the appropriate reference below.

    References
    ----------
    .. [1] Shen, Y., and Bax, Ad. "SPARTA+: a modest improvement in empirical
       NMR chemical shift prediction by means of an artificial neural network."
       J. Biomol. NMR, 48, 13-22 (2010)
    """
    pd = import_('pandas')
    binary = find_executable(SPARTA_PLUS)
    if binary is None:
        raise OSError('External command not found. Looked for %s in PATH. `chemical_shifts_spartaplus` requires the external program SPARTA+, available at http://spin.niddk.nih.gov/bax/software/SPARTA+/' % ', '.join(SPARTA_PLUS))

    names = ["resSeq", "resName", "name", "SS_SHIFT", "SHIFT", "RC_SHIFT", "HM_SHIFT", "EF_SHIFT", "SIGMA"]

    with enter_temp_directory():
        for i in range(trj.n_frames):
            trj[i].save("./trj%d.pdb" % i)

        subprocess.check_call([binary, '-in'] + ["trj{}.pdb".format(i) for i in range(trj.n_frames)]
                              + ['-out', 'trj0_pred.tab'])

        lines_to_skip = _get_lines_to_skip("trj0_pred.tab")

        results = []
        for i in range(trj.n_frames):
            d = pd.read_table("./trj%d_pred.tab" % i, names=names, header=None, sep="\s+", skiprows=lines_to_skip)
            d["frame"] = i
            results.append(d)

    results = pd.concat(results)

    if rename_HN:
        results.name[results.name == "HN"] = "H"

    if LooseVersion(pd.__version__) < LooseVersion('0.14.0'):
        results = results.pivot_table(rows=["resSeq", "name"], cols="frame", values="SHIFT")
    else:
        results = results.pivot_table(index=["resSeq", "name"], columns="frame", values="SHIFT")

    return results


def reindex_dataframe_by_atoms(trj, frame):
    """Reindex chemical shift output to use atom number (serial) indexing.

    Parameters
    ----------
    trj : Trajectory
        Trajectory to predict shifts for.
    frame : pandas.DataFrame
        Dataframe containing results, with index consisting of
        (resSeq, atom_name) pairs and columns for each frame in trj.

    Returns
    -------
    new_frame : pandas.DataFrame
        Dataframe containing results, with index consisting of atom
        indices (AKA the 'serial' entry in a PDB).  Columns correspond to
        each frame in trj.

    Notes
    -----
    Be aware that this function may DROP predictions if the atom naming
    is different between the input trajectory and the output of various
    chemical shift prediction tools.
    """

    top, bonds = trj.top.to_dataframe()
    top["serial"] = top.index
    top = top.set_index(["resSeq", "name"])

    new_frame = frame.copy()

    new_frame["serial"] = top.ix[new_frame.index].serial
    new_frame = new_frame.dropna().reset_index().set_index("serial").drop(["resSeq", "name"], axis=1)

    return new_frame



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

"""OpenMM Reporter for saving the state of a molecular dynamics simulation
through time in the AMBER NetCDF format
"""


from __future__ import print_function, division
from ..formats.dcd import DCDTrajectoryFile
from .basereporter import _BaseReporter


class DCDReporter(_BaseReporter):
    """DCDReporter stores a molecular dynamics trajectory in the CHARMM / NAMD
    DCD Format

    Parameters
    ----------
    file : str, or DCDTrajectoryFile
        Either an open DCDTrajectoryFile object to write to, or a string
        specifying the filename of a new DCD file to save the trajectory to.
    reportInterval : int
        The interval (in time steps) at which to write frames.
    atomSubset : array_like, default=None
        Only write a subset of the atoms, with these (zero based) indices
        to the file. If None, *all* of the atoms will be written to disk.

    Examples
    --------
    >>> simulation = Simulation(topology, system, integrator)
    >>> dcd_reporter = DCDReporter('traj.dcd', 100)
    >>> simulation.reporters.append(dcd_reporter)
    >>> simulation.step(10000)

    >>> traj = mdtraj.trajectory.load('traj.dcd')
    """
    @property
    def backend(self):
        return DCDTrajectoryFile

    def __init__(self, file, reportInterval, atomSubset=None):
        super(DCDReporter, self).__init__(file, reportInterval,
            coordinates=True, time=False, cell=True, potentialEnergy=False,
            kineticEnergy=False, temperature=False, velocities=False,
            atomSubset=atomSubset)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Alexander Yang
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


##############################################################################
# Imports
##############################################################################

from __future__ import print_function, division

import numpy as np
import os


from mdtraj.core.topology import Topology
from mdtraj.core.element import virtual_site
from mdtraj.utils import (ensure_type, cast_indices, 
        lengths_and_angles_to_tilt_factors)
from mdtraj.formats.registry import FormatRegistry
from mdtraj.utils.six import string_types

__all__ = ['load_gsd', 'write_gsd', 'load_gsd_topology']


@FormatRegistry.register_loader('.gsd')
def load_gsd(filename, top=None, start=None, n_frames=None, stride=None, 
        atom_indices=None, frame=None):
    """Load a GSD trajectory file.

    Parameters
    -----------
    filename : path-like
        Path of GSD trajectory file.
    top : {path-like, Trajectory, Topology}, None
        A pdb file, a trajectory, or a topology to supply topology information
        If None, topology information will be parsed from the GSD file
    start : int, None
        First frame to convert
    n_frames : int, None
        Number of frames after `start` to convert
    stride : int
        Read only every stride-th frame.   
    atom_indices : array_like, optional
        If not none, then read only a subset of the atoms coordinates from the
        file.
    frame : int, optional
        Use this option to load only a single frame from a trajectory on disk.
        If frame is None, the default, the entire trajectory will be loaded.
        If supplied, ``stride`` will be ignored.

    Returns
    -------
    trajectory : md.Trajectory
        The resulting trajectory, as an md.Trajectory object.

    """
    from mdtraj.core.trajectory import Trajectory, _parse_topology
    import gsd.hoomd

    if not isinstance(filename, (string_types, os.PathLike)):
        raise TypeError('filename must be of type path-like for load_gsd. '
                        'you supplied %s'.format(type(filename)))

    if top is not None:
        topology = _parse_topology(top)
    else:
        topology = load_gsd_topology(filename)
    atom_indices = cast_indices(atom_indices)

    with gsd.hoomd.open(filename, 'rb') as f:
        if frame is not None:
            xyz, vectors, time = read_snapshot(frame, f[frame], 
                    topology, atom_indices=atom_indices)
            t = Trajectory(xyz=np.array(xyz), topology=topology, 
                    time=np.array([time]))
            t.unitcell_vectors = np.reshape(vectors, (-1,3,3))
            return t

        else:
            return hoomdtraj_to_traj(f, topology, start=start, n_frames=n_frames,
                    stride=stride, atom_indices=atom_indices)

def load_gsd_topology(filename, frame=0):
    """ Create an MDTraj.Topology from a GSD file 
    
    Parameters
    ----------
    filename : path-like
        Path of GSD trajectory file.
    frame : int, 0 
        Frame of GSD file to parse topology

    Returns
    -------
    top : mdtraj.Topology

    Notes
    -----
    GSD files support systems with variable topologies.
    For compatibility with MDTraj, only the topology from GSD frame 0 is
    used to construct the MDTraj topology.
    """
    import gsd.hoomd
    with gsd.hoomd.open(filename, 'rb') as gsdfile:
        top = Topology()
        generic_chain = top.add_chain()
        generic_residue = top.add_residue('A', generic_chain)
        all_particle_types = gsdfile[frame].particles.types
        for particle_type_id in gsdfile[frame].particles.typeid:
            top.add_atom(all_particle_types[particle_type_id], virtual_site,
                    generic_residue)

        for bond in gsdfile[frame].bonds.group:
            atom1, atom2 = bond[0], bond[1]
            top.add_bond(top.atom(atom1), top.atom(atom2))

    return top

def hoomdtraj_to_traj(f, topology, start=None, n_frames=None, 
        stride=None, atom_indices=None):
    """ Convert HOOMDTrajectory to MDtraj Trajectory 
   
    Parameters
    ----------
    f : gsd.hoomd.HOOMDTrajectory object
    topology : mdtraj.Topology
    start : int, None
        First frame to convert
    n_frames : int, None
        Number of frames after `start` to convert
    stride : int
        Read only every stride-th frame.   
    atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates
            from the file.

    Returns
    --------
    traj : mdtraj.Trajectory object
    
    """
    from mdtraj.core.trajectory import Trajectory
    if start is None:
        start = 0
    if n_frames is None:
        n_frames = len(f) - start
    if stride is None:
        stride = 1

    all_coords, all_times, all_vectors = [], [], []
    for i, snapshot in enumerate(f[start : start+n_frames : stride], start=start):
        xyz, box_vectors, time = read_snapshot(i, snapshot, topology, 
                atom_indices=atom_indices)
        all_coords.append(xyz)
        all_vectors.append(box_vectors)
        all_times.append(time)

    all_coords = np.array(all_coords)
    all_vectors = np.array(all_vectors)
    all_times = np.array(all_times)
    if len(all_coords) == 0:
        return Trajectory(xyz=np.zeros((0, topology.n_atoms, 3)), 
                topology=topology)

    t = Trajectory(xyz=all_coords, topology=topology, time=all_times)
    t.unitcell_vectors = all_vectors
    return t

def read_snapshot(frame, snapshot, topology, atom_indices=None):
    """ Parse relevant information from a single HOOMD snapshot (frame) 
    
    Parameters
    ----------
    frame : int
        Frame index to read
    snapshot : gsd.hoomd.Snapshot
    topology : mdtraj.Topology
    atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates
            from the file.

    Returns
    --------
    xyz : np.ndarray, (n, 3)
    box_vectors : list, (3, 3)
    time : int
        Step of hoomd snapshot

    Notes
    -----
    If the GSD file has a variable topology compared to the
    given topology, an IOError will be raised.
    
    """

    _check_topology(frame, snapshot, topology)
    xyz = snapshot.particles.position

    if atom_indices is not None:
        xyz = xyz[atom_indices]
    lx, ly, lz, xy, xz, yz = snapshot.configuration.box
    box_vectors=[[lx, xy*ly, xz*lz],
                            [0.0, ly, yz*lz],
                            [0.0, 0.0,lz]]
    time = snapshot.configuration.step

    return xyz, box_vectors, time

def write_gsd(filename, xyz, top, cell_lengths=None, cell_angles=None):
    """Write one or more frames of data to a gsd file.

    Parameters
    ----------
    filename : path-like
    xyz : np.ndarray, shape=(n_frames, n_atoms, 3)
        The cartesian coordinates of the atoms to write. 
    top : mdtraj.Topology
    cell_lengths : np.ndarray, dtype=np.double, shape=(n_frames, 3)
        The lengths (a,b,c) of the unit cell for each frame. 
    cell_angles : np.ndarray, dtype=np.double, shape=(n_frames, 3)
        The angles (\alpha, \beta, \gamma) defining the unit cell for
        each frame. (Units of degrees).
    """
    import gsd.hoomd
    xyz = ensure_type(xyz, np.float32, 3, 'xyz', can_be_none=False,
            shape=(None, None, 3), warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True)
    cell_lengths = ensure_type(cell_lengths, np.float32, 2, 'cell_lengths',
            can_be_none=False, shape=(len(xyz), 3), warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True)
    if cell_angles is None:
        cell_angles = np.empty_like(cell_lengths)
        cell_angles.fill(90)
    cell_angles = ensure_type(cell_angles, np.float32, 2, 'cell_angles',
            can_be_none=False, shape=(len(xyz), 3), warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True)
    unique_types = list(set(a.name for a in top.atoms)) # Find unique atomtypes

    types = {a:i for i, a in enumerate(unique_types)} # Map str(atomtype) to index
    atomtype_ids = np.array([types[a.name] for a in top.atoms], dtype=np.int32)

    unique_bond_types = []
    bondtype_ids = []
    bond_groups = []
    if top.n_bonds > 0:
        unique_bond_types, bondtype_ids, bond_groups = _process_bonds(top)

    with gsd.hoomd.open(filename, 'wb') as hoomd_traj:
        for i, coords in enumerate(xyz):
            gsd_frame = gsd.hoomd.Snapshot()
            gsd_frame.particles.N = top.n_atoms
            gsd_frame.particles.position = coords 
            gsd_frame.particles.types = unique_types
            gsd_frame.particles.typeid = atomtype_ids
            gsd_frame.bonds.N = top.n_bonds
            gsd_frame.bonds.types = unique_bond_types
            gsd_frame.bonds.typeid = bondtype_ids
            gsd_frame.bonds.group = bond_groups

            gsd_frame.configuration.box = lengths_and_angles_to_tilt_factors(
                    cell_lengths[i][0], cell_lengths[i][1], cell_lengths[i][2],
                    cell_angles[i][0], cell_angles[i][1], cell_angles[i][2])
            hoomd_traj.append(gsd_frame)

def _process_bonds(top):
    """ Identify and relate unique bondtypes to an index """
    sorted_bond_types = [tuple(sorted([b[0].name, b[1].name])) for b in top.bonds]
    unique_bond_types = list(set("{}-{}".format(b[0], b[1]) 
            for b in sorted_bond_types))
    bondtypes = {a:i for i, a in enumerate(unique_bond_types)}
    bondtype_ids = [bondtypes["{}-{}".format(key[0], key[1])] 
        for key in sorted_bond_types]
    bond_groups = [(b[0].index, b[1].index) for b in top.bonds]

    return (unique_bond_types, bondtype_ids, bond_groups)

def _check_topology(frame, snapshot, topology):
    """ Verify snapshot topology matches given MDTraj topology 
    
    Notes
    -----
    Only looks for N particles and N bonds for speed purposes"""
    error_msg = ("GSD frame {} ".format(frame) + 
            "has inconsistent topology compared to given topology, " + 
            "this is unsupported in MDTraj.")
    if (snapshot.particles.N != topology.n_atoms or
            snapshot.bonds.N != topology.n_bonds):
        raise IOError(error_msg)



from __future__ import absolute_import

from .arc import ArcTrajectoryFile
from .dcd import DCDTrajectoryFile
from .binpos import BINPOSTrajectoryFile
from .xtc import XTCTrajectoryFile
from .trr import TRRTrajectoryFile
from .hdf5 import HDF5TrajectoryFile
from .netcdf import NetCDFTrajectoryFile
from .pdb import PDBTrajectoryFile
from .lh5 import LH5TrajectoryFile
from .mdcrd import MDCRDTrajectoryFile
from .amberrst import AmberRestartFile, AmberNetCDFRestartFile
from .lammpstrj import LAMMPSTrajectoryFile
from .dtr import DTRTrajectoryFile
from .gro import GroTrajectoryFile
from .xyzfile import XYZTrajectoryFile
from .tng import TNGTrajectoryFile



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

"""OpenMM Reporter for saving the positions of a molecular dynamics simulation
in the HDF5 format.
"""

##############################################################################
# Imports
##############################################################################

from __future__ import print_function, division
from mdtraj.formats.hdf5 import HDF5TrajectoryFile
from mdtraj.reporters.basereporter import _BaseReporter

##############################################################################
# Classes
##############################################################################


class HDF5Reporter(_BaseReporter):
    """HDF5Reporter stores a molecular dynamics trajectory in the HDF5 format.

    This object supports saving all kinds of information from the simulation --
    more than any other trajectory format. In addition to all of the options,
    the topology of the system will also (of course) be stored in the file. All
    of the information is compressed, so the size of the file is not much
    different than DCD, despite the added flexibility.

    Parameters
    ----------
    file : str, or HDF5TrajectoryFile
        Either an open HDF5TrajecoryFile object to write to, or a string
        specifying the filename of a new HDF5 file to save the trajectory to.
    reportInterval : int
        The interval (in time steps) at which to write frames.
    coordinates : bool
        Whether to write the coordinates to the file.
    time : bool
        Whether to write the current time to the file.
    cell : bool
        Whether to write the current unit cell dimensions to the file.
    potentialEnergy : bool
        Whether to write the potential energy to the file.
    kineticEnergy : bool
        Whether to write the kinetic energy to the file.
    temperature : bool
        Whether to write the instantaneous temperature to the file.
    velocities : bool
        Whether to write the velocities to the file.
    atomSubset : array_like, default=None
        Only write a subset of the atoms, with these (zero based) indices
        to the file. If None, *all* of the atoms will be written to disk.
    enforcePeriodicBox: bool or None
        Specifies whether particle positions should be translated so the
        center of every molecule lies in the same periodic box. If None
        (the default), it will automatically decide whether to translate
        molecules based on whether the system being simulated uses periodic
        boundary conditions.

    Notes
    -----
    If you use the ``atomSubset`` option to write only a subset of the atoms
    to disk, the ``kineticEnergy``, ``potentialEnergy``, and ``temperature``
    fields will not change. They will still refer to the energy and temperature
    of the *whole* system, and are not "subsetted" to only include the energy
    of your subsystem.

    Examples
    --------
    >>> simulation = Simulation(topology, system, integrator)
    >>> h5_reporter = HDF5Reporter('traj.h5', 100)
    >>> simulation.reporters.append(h5_reporter)
    >>> simulation.step(10000)

    >>> traj = mdtraj.trajectory.load('traj.lh5')
    """
    @property
    def backend(self):
        return HDF5TrajectoryFile

    def __init__(self, file, reportInterval, coordinates=True, time=True,
                 cell=True, potentialEnergy=True, kineticEnergy=True,
                 temperature=True, velocities=False, atomSubset=None,
                 enforcePeriodicBox=None):
        """Create a HDF5Reporter.
        """
        super(HDF5Reporter, self).__init__(file, reportInterval,
            coordinates, time, cell, potentialEnergy, kineticEnergy,
            temperature, velocities, atomSubset,
            enforcePeriodicBox)



"""
Registry for trajectory file formats, so that the appropriate file
object and loader can be resolved based on the filename extension.
"""

class FormatRegistry(object):
    """Registry for trajectory file objects.

    Examples
    --------
    >>> @FormatRegistry.register_loader('.xyz')
    >>> def load_xyz(filename):
        return Trajectory(...)
    >>>
    >>> print FormatRegistry.loaders['.xyz']
    <function load_xyz at 0x1004a15f0>
    """
    loaders = {}
    fileobjects = {}

    @classmethod
    def register_loader(cls, extension):
        def decorator(f):
            cls.loaders[extension] = f
            return f
        return decorator

    @classmethod
    def register_fileobject(cls, extension):
        def decorator(f):
            cls.fileobjects[extension] = f
            return f
        return decorator

# Make a single instance of this class, and then
# get rid of the class object. This should be
# treated as a singleton
FormatRegistry = FormatRegistry()

# for backward compatibility reasons, we have this alias:
_FormatRegistry = FormatRegistry



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Kyle A Beauchamp
# Contributors: Robert McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

"""Pure python code to calculate the radius of gyration of frames in a
trajectory
"""
##############################################################################
# Imports
##############################################################################

from __future__ import print_function, division
import numpy as np

__all__ = ['compute_rg']

##############################################################################
# Functions
##############################################################################


def _compute_rg_xyz(xyz, masses=None):
    """Compute the radius of gyration for every frame.

    Parameters
    ----------
    xyz : ndarray
        xyz coordinates
    masses : ndarray
        Transition matrix

    Returns
    -------
    Rg : ndarray
        Rg for every frame

    Notes
    -----
    If masses are none, assumes equal masses.
    """
    traj_length, num_atoms, num_dims = xyz.shape
    if not num_dims == 3:
        raise ValueError("What did you pass me?")
    if not xyz.dtype == np.float32:
        xyz = np.float32(xyz)
    if masses is None:
        masses = np.ones(num_atoms)

    weights = masses / masses.sum()

    mu = xyz.mean(1)
    centered = (xyz.transpose((1, 0, 2)) - mu).transpose((1, 0, 2))
    squared_dists = (centered ** 2).sum(2)
    Rg = (squared_dists * weights).sum(1) ** 0.5

    return Rg


def compute_rg(traj, masses=None):
    """Compute the radius of gyration for every frame.

    Parameters
    ----------
    traj : Trajectory

    masses : ndarray, optional
        array of atom masses.

    Returns
    -------
    Rg : ndarray
        Rg for every frame

    Notes
    -----
    If masses are none, assumes equal masses.
    """
    return _compute_rg_xyz(traj.xyz, masses=masses)



from __future__ import print_function, division
import os
import time
import shutil
import tempfile
import contextlib

__all__ = ["timing", "enter_temp_directory"]


class timing(object):
    """A timing context manager

    Examples
    --------
    >>> long_function = lambda : None
    >>> with timing('long_function'):
    ...     long_function()
    long_function: 0.000 seconds
    """
    def __init__(self, name='block'):
        self.name = name
        self.time = 0
        self.start = None
        self.end = None
    
    def __enter__(self):
        self.start = time.time()
        return self
    
    def __exit__(self, ty, val, tb):
        self.end = time.time()
        self.time = self.end - self.start
        print("%s: %0.3f seconds" % (self.name, self.time))
        return False


@contextlib.contextmanager
def enter_temp_directory():
    """Create and enter a temporary directory; used as context manager."""
    temp_dir = tempfile.mkdtemp()
    cwd = os.getcwd()
    os.chdir(temp_dir)
    yield
    os.chdir(cwd)
    shutil.rmtree(temp_dir)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Tim Moore
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################
from xml.etree import cElementTree

import numpy as np

from mdtraj.formats.registry import FormatRegistry
from mdtraj.utils import ilen, import_, ensure_type
from mdtraj.core.element import virtual_site


__all__ = ['load_hoomdxml']


@FormatRegistry.register_loader('.hoomdxml')
def load_hoomdxml(filename, top=None):
    """Load a single conformation from an HOOMD-Blue XML file.

    For more information on this file format, see:
    http://codeblue.umich.edu/hoomd-blue/doc/page_xml_file_format.html
    Notably, all node names and attributes are in all lower case.
    HOOMD-Blue does not contain residue and chain information explicitly. 
    For this reason, chains will be found by looping over all the bonds and 
    finding what is bonded to what. 
    Each chain consisists of exactly one residue. 

    Parameters
    ----------
    filename : path-like
        The path on disk to the XML file
    top : None
        This argumet is ignored

    Returns
    -------
    trajectory : md.Trajectory
        The resulting trajectory, as an md.Trajectory object, with corresponding 
        Topology.

    Notes
    -----
    This function requires the NetworkX python package.
    """
    from mdtraj.core.trajectory import Trajectory
    from mdtraj.core.topology import Topology
    topology = Topology()
    tree = cElementTree.parse(filename)
    config = tree.getroot().find('configuration')
    position = config.find('position')
    bond = config.find('bond')
    atom_type = config.find('type')  # MDTraj calls this "name"

    box = config.find('box')
    box.attrib = dict((key.lower(), val) for key, val in box.attrib.items())
    # be generous for case of box attributes
    lx = float(box.attrib['lx'])
    ly = float(box.attrib['ly'])
    lz = float(box.attrib['lz'])
    try:
        xy = float(box.attrib['xy'])
        xz = float(box.attrib['xz'])
        yz = float(box.attrib['yz'])
    except (ValueError, KeyError):
        xy = 0.0
        xz = 0.0
        yz = 0.0
    unitcell_vectors = np.array([[[lx,  xy*ly, xz*lz],
                                  [0.0, ly,    yz*lz],
                                  [0.0, 0.0,   lz   ]]])

    positions, types = [], {}
    for pos in position.text.splitlines()[1:]:
        positions.append((float(pos.split()[0]),
                          float(pos.split()[1]),
                          float(pos.split()[2])))

    for idx, atom_name in enumerate(atom_type.text.splitlines()[1:]):
        types[idx] = str(atom_name.split()[0])
    if len(types) != len(positions):
        raise ValueError('Different number of types and positions in xml file')

    # ignore the bond type
    if hasattr(bond, 'text'):
        bonds = [(int(b.split()[1]), int(b.split()[2])) for b in bond.text.splitlines()[1:]]
        chains = _find_chains(bonds)
    else:
        chains = []
        bonds = []

    # Relate the first index in the bonded-group to mdtraj.Residue
    bonded_to_residue = {}
    for i, _ in enumerate(types):
        bonded_group = _in_chain(chains, i)
        if bonded_group is not None:
            if bonded_group[0] not in bonded_to_residue:
                t_chain = topology.add_chain()
                t_residue = topology.add_residue('A', t_chain)
                bonded_to_residue[bonded_group[0]] = t_residue
            topology.add_atom(types[i], virtual_site, 
                    bonded_to_residue[bonded_group[0]])
        if bonded_group is None:
            t_chain = topology.add_chain()
            t_residue = topology.add_residue('A', t_chain)
            topology.add_atom(types[i], virtual_site, t_residue)

    for bond in bonds:
        atom1, atom2 = bond[0], bond[1]
        topology.add_bond(topology.atom(atom1), topology.atom(atom2))

    traj = Trajectory(xyz=np.array(positions), topology=topology)
    traj.unitcell_vectors = unitcell_vectors

    return traj

def _find_chains(bond_list):
    """Given a set of bonds, find unique molecules, with the assumption that
    there are no bonds between separate chains (i.e., only INTRAmolecular
    bonds), which also implies that each atom can be in exactly one chain.
    
    Parameters
    ----------
    bond_list : list of (int, int)
        The list of bonds

    Returns
    _______
    chains : list of list of int
        List of atoms in each chain

    Notes
    -----
    This function requires the NetworkX python package.
    """
    nx = import_('networkx')
    chains = []
    bond_list = np.asarray(bond_list)
    molecules = nx.Graph()
    molecules.add_nodes_from(set(bond_list.flatten()))
    molecules.add_edges_from(bond_list)
    return [sorted(x) for x in list(nx.connected_components(molecules))]

def _in_chain(chains, atom_index):
    """Check if an item is in a list of lists"""
    for chain in chains:
        if atom_index in chain:
            return chain
    return None
